@@ -52,6 +52,19 @@ static const struct sc_asn1_entry c_asn1_prk_rsa_attr[] = {
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
+static const struct sc_asn1_entry c_asn1_gostr3410key_attr[] = {
+	{ "value",         SC_ASN1_PATH, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ "params_r3410",  SC_ASN1_INTEGER, SC_ASN1_TAG_INTEGER, 0, NULL, NULL },
+	{ "params_r3411",  SC_ASN1_INTEGER, SC_ASN1_TAG_INTEGER, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "params_28147",  SC_ASN1_INTEGER, SC_ASN1_TAG_INTEGER, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
+};
+
+static const struct sc_asn1_entry c_asn1_prk_gostr3410_attr[] = {
+	{ "privateGOSTR3410KeyAttributes", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
+};
+
 static const struct sc_asn1_entry c_asn1_dsakey_i_p_attr[] = {
 	{ "path",	SC_ASN1_PATH, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
@@ -76,8 +89,9 @@ static const struct sc_asn1_entry c_asn1_prk_dsa_attr[] = {
 static const struct sc_asn1_entry c_asn1_prkey[] = {
 	{ "privateRSAKey", SC_ASN1_PKCS15_OBJECT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "privateDSAKey", SC_ASN1_PKCS15_OBJECT,  2 | SC_ASN1_CTX | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "privateGOSTR3410Key", SC_ASN1_PKCS15_OBJECT, 3 | SC_ASN1_CTX | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
-};		
+};
 
 int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
 				 struct sc_pkcs15_object *obj,
@@ -85,7 +99,8 @@ int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
 {
         sc_context_t *ctx = p15card->card->ctx;
         struct sc_pkcs15_prkey_info info;
-	int r;
+	int r, gostr3410_params[3];
+	struct sc_pkcs15_keyinfo_gostparams *keyinfo_gostparams;
 	size_t usage_len = sizeof(info.usage);
 	size_t af_len = sizeof(info.access_flags);
 	struct sc_asn1_entry asn1_com_key_attr[6], asn1_com_prkey_attr[1];
@@ -93,30 +108,37 @@ int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
 	struct sc_asn1_entry asn1_dsakey_attr[2], asn1_prk_dsa_attr[2],
 			asn1_dsakey_i_p_attr[2],
 			asn1_dsakey_value_attr[3];
-	struct sc_asn1_entry asn1_prkey[3];
+	struct sc_asn1_entry asn1_gostr3410key_attr[5], asn1_prk_gostr3410_attr[2];
+	struct sc_asn1_entry asn1_prkey[4];
 	struct sc_asn1_pkcs15_object rsa_prkey_obj = { obj, asn1_com_key_attr,
 						       asn1_com_prkey_attr, asn1_prk_rsa_attr };
 	struct sc_asn1_pkcs15_object dsa_prkey_obj = { obj, asn1_com_key_attr,
 						       asn1_com_prkey_attr, asn1_prk_dsa_attr };
+	struct sc_asn1_pkcs15_object gostr3410_prkey_obj =  { obj, asn1_com_key_attr,
+						       asn1_com_prkey_attr,
+						       asn1_prk_gostr3410_attr };
 
-        sc_copy_asn1_entry(c_asn1_prkey, asn1_prkey);
+	sc_copy_asn1_entry(c_asn1_prkey, asn1_prkey);
 
-        sc_copy_asn1_entry(c_asn1_prk_rsa_attr, asn1_prk_rsa_attr);
-        sc_copy_asn1_entry(c_asn1_rsakey_attr, asn1_rsakey_attr);
-        sc_copy_asn1_entry(c_asn1_prk_dsa_attr, asn1_prk_dsa_attr);
-        sc_copy_asn1_entry(c_asn1_dsakey_attr, asn1_dsakey_attr);
+	sc_copy_asn1_entry(c_asn1_prk_rsa_attr, asn1_prk_rsa_attr);
+	sc_copy_asn1_entry(c_asn1_rsakey_attr, asn1_rsakey_attr);
+	sc_copy_asn1_entry(c_asn1_prk_dsa_attr, asn1_prk_dsa_attr);
+	sc_copy_asn1_entry(c_asn1_dsakey_attr, asn1_dsakey_attr);
 	sc_copy_asn1_entry(c_asn1_dsakey_value_attr, asn1_dsakey_value_attr);
-	sc_copy_asn1_entry(c_asn1_dsakey_i_p_attr,
-					asn1_dsakey_i_p_attr);
+	sc_copy_asn1_entry(c_asn1_dsakey_i_p_attr, asn1_dsakey_i_p_attr);
+	sc_copy_asn1_entry(c_asn1_prk_gostr3410_attr, asn1_prk_gostr3410_attr);
+	sc_copy_asn1_entry(c_asn1_gostr3410key_attr, asn1_gostr3410key_attr);
 
-        sc_copy_asn1_entry(c_asn1_com_prkey_attr, asn1_com_prkey_attr);
-        sc_copy_asn1_entry(c_asn1_com_key_attr, asn1_com_key_attr);
+	sc_copy_asn1_entry(c_asn1_com_prkey_attr, asn1_com_prkey_attr);
+	sc_copy_asn1_entry(c_asn1_com_key_attr, asn1_com_key_attr);
 
 	sc_format_asn1_entry(asn1_prkey + 0, &rsa_prkey_obj, NULL, 0);
 	sc_format_asn1_entry(asn1_prkey + 1, &dsa_prkey_obj, NULL, 0);
+	sc_format_asn1_entry(asn1_prkey + 2, &gostr3410_prkey_obj, NULL, 0);
 
 	sc_format_asn1_entry(asn1_prk_rsa_attr + 0, asn1_rsakey_attr, NULL, 0);
 	sc_format_asn1_entry(asn1_prk_dsa_attr + 0, asn1_dsakey_attr, NULL, 0);
+	sc_format_asn1_entry(asn1_prk_gostr3410_attr + 0, asn1_gostr3410key_attr, NULL, 0);
 
 	sc_format_asn1_entry(asn1_rsakey_attr + 0, &info.path, NULL, 0);
 	sc_format_asn1_entry(asn1_rsakey_attr + 1, &info.modulus_length, NULL, 0);
@@ -126,6 +148,10 @@ int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
 	sc_format_asn1_entry(asn1_dsakey_value_attr + 1, asn1_dsakey_i_p_attr, NULL, 0);
 	sc_format_asn1_entry(asn1_dsakey_i_p_attr + 0, &info.path, NULL, 0);
 
+	sc_format_asn1_entry(asn1_gostr3410key_attr + 0, &info.path, NULL, 0);
+	sc_format_asn1_entry(asn1_gostr3410key_attr + 1, &gostr3410_params[0], NULL, 0);
+	sc_format_asn1_entry(asn1_gostr3410key_attr + 2, &gostr3410_params[1], NULL, 0);
+	sc_format_asn1_entry(asn1_gostr3410key_attr + 3, &gostr3410_params[2], NULL, 0);
 
 	sc_format_asn1_entry(asn1_com_key_attr + 0, &info.id, NULL, 0);
 	sc_format_asn1_entry(asn1_com_key_attr + 1, &info.usage, &usage_len, 0);
@@ -137,6 +163,7 @@ int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
         memset(&info, 0, sizeof(info));
 	info.key_reference = -1;
 	info.native = 1;
+	memset(gostr3410_params, 0, sizeof(gostr3410_params));
 
 	r = sc_asn1_decode_choice(ctx, asn1_prkey, *buf, *buflen, buf, buflen);
 	if (r == SC_ERROR_ASN1_END_OF_CONTENTS)
@@ -149,16 +176,46 @@ int sc_pkcs15_decode_prkdf_entry(struct sc_pkcs15_card *p15card,
 		/* If the value was indirect-protected, mark the path */
 		if (asn1_dsakey_i_p_attr[0].flags & SC_ASN1_PRESENT)
 			info.path.type = SC_PATH_TYPE_PATH_PROT;
+	} else if (asn1_prkey[2].flags & SC_ASN1_PRESENT) {
+		obj->type = SC_PKCS15_TYPE_PRKEY_GOSTR3410;
+		assert(info.modulus_length == 0);
+		info.modulus_length = SC_PKCS15_GOSTR3410_KEYSIZE;
+		assert(info.params_len == 0);
+		info.params_len = sizeof(struct sc_pkcs15_keyinfo_gostparams);
+		info.params = malloc(info.params_len);
+		if (info.params == NULL)
+			SC_FUNC_RETURN(ctx, 0, SC_ERROR_OUT_OF_MEMORY);
+		assert(sizeof(*keyinfo_gostparams) == info.params_len);
+		keyinfo_gostparams = info.params;
+		keyinfo_gostparams->gostr3410 = gostr3410_params[0];
+		keyinfo_gostparams->gostr3411 = gostr3410_params[1];
+		keyinfo_gostparams->gost28147 = gostr3410_params[2];
 	} else {
-		sc_error(ctx, "Neither RSA or DSA key in PrKDF entry.\n");
+		sc_error(ctx, "Neither RSA or DSA or GOSTR3410 key in PrKDF entry.\n");
 		SC_FUNC_RETURN(ctx, 0, SC_ERROR_INVALID_ASN1_OBJECT);
 	}
 	r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &info.path);
-	if (r < 0)
+	if (r < 0) {
+		if (info.params)
+			free(info.params);
 		return r;
+	}
+
+        /* OpenSC 0.11.4 and older encoded "keyReference" as a negative
+           value. Fixed in 0.11.5 we need to add a hack, so old cards
+           continue to work. */
+        if (p15card->flags & SC_PKCS15_CARD_FLAG_FIX_INTEGERS) {
+                if (info.key_reference < -1) {
+                        info.key_reference += 256;
+                }
+        }
+
 	obj->data = malloc(sizeof(info));
-	if (obj->data == NULL)
+	if (obj->data == NULL) {
+		if (info.params)
+			free(info.params);
 		SC_FUNC_RETURN(ctx, 0, SC_ERROR_OUT_OF_MEMORY);
+	}
 	memcpy(obj->data, &info, sizeof(info));
 
 	return 0;
@@ -173,27 +230,34 @@ int sc_pkcs15_encode_prkdf_entry(sc_context_t *ctx,
 	struct sc_asn1_entry asn1_dsakey_attr[2], asn1_prk_dsa_attr[2],
 			asn1_dsakey_value_attr[3],
 			asn1_dsakey_i_p_attr[2];
-	struct sc_asn1_entry asn1_prkey[3];
+	struct sc_asn1_entry asn1_gostr3410key_attr[5], asn1_prk_gostr3410_attr[2];
+	struct sc_asn1_entry asn1_prkey[4];
 	struct sc_asn1_pkcs15_object rsa_prkey_obj = { (struct sc_pkcs15_object *) obj, asn1_com_key_attr,
 						       asn1_com_prkey_attr, asn1_prk_rsa_attr };
 	struct sc_asn1_pkcs15_object dsa_prkey_obj = { (struct sc_pkcs15_object *) obj, asn1_com_key_attr,
 						       asn1_com_prkey_attr, asn1_prk_dsa_attr };
+	struct sc_asn1_pkcs15_object gostr3410_prkey_obj =  { (struct sc_pkcs15_object *) obj,
+						       asn1_com_key_attr, asn1_com_prkey_attr,
+						       asn1_prk_gostr3410_attr };
 	struct sc_pkcs15_prkey_info *prkey =
                 (struct sc_pkcs15_prkey_info *) obj->data;
+	struct sc_pkcs15_keyinfo_gostparams *keyinfo_gostparams;
 	int r;
 	size_t af_len, usage_len;
 
-        sc_copy_asn1_entry(c_asn1_prkey, asn1_prkey);
+	sc_copy_asn1_entry(c_asn1_prkey, asn1_prkey);
 
-        sc_copy_asn1_entry(c_asn1_prk_rsa_attr, asn1_prk_rsa_attr);
-        sc_copy_asn1_entry(c_asn1_rsakey_attr, asn1_rsakey_attr);
-        sc_copy_asn1_entry(c_asn1_prk_dsa_attr, asn1_prk_dsa_attr);
-        sc_copy_asn1_entry(c_asn1_dsakey_attr, asn1_dsakey_attr);
+	sc_copy_asn1_entry(c_asn1_prk_rsa_attr, asn1_prk_rsa_attr);
+	sc_copy_asn1_entry(c_asn1_rsakey_attr, asn1_rsakey_attr);
+	sc_copy_asn1_entry(c_asn1_prk_dsa_attr, asn1_prk_dsa_attr);
+	sc_copy_asn1_entry(c_asn1_dsakey_attr, asn1_dsakey_attr);
 	sc_copy_asn1_entry(c_asn1_dsakey_value_attr, asn1_dsakey_value_attr);
 	sc_copy_asn1_entry(c_asn1_dsakey_i_p_attr, asn1_dsakey_i_p_attr);
+	sc_copy_asn1_entry(c_asn1_prk_gostr3410_attr, asn1_prk_gostr3410_attr);
+	sc_copy_asn1_entry(c_asn1_gostr3410key_attr, asn1_gostr3410key_attr);
 
-        sc_copy_asn1_entry(c_asn1_com_prkey_attr, asn1_com_prkey_attr);
-        sc_copy_asn1_entry(c_asn1_com_key_attr, asn1_com_key_attr);
+	sc_copy_asn1_entry(c_asn1_com_prkey_attr, asn1_com_prkey_attr);
+	sc_copy_asn1_entry(c_asn1_com_key_attr, asn1_com_key_attr);
 
 	switch (obj->type) {
 	case SC_PKCS15_TYPE_PRKEY_RSA:
@@ -217,6 +281,21 @@ int sc_pkcs15_encode_prkdf_entry(sc_context_t *ctx,
 					&prkey->path, NULL, 1);
 		}
                 break;
+	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
+		sc_format_asn1_entry(asn1_prkey + 2, &gostr3410_prkey_obj, NULL, 1);
+		sc_format_asn1_entry(asn1_prk_gostr3410_attr + 0, asn1_gostr3410key_attr, NULL, 1);
+		sc_format_asn1_entry(asn1_gostr3410key_attr + 0, &prkey->path, NULL, 1);
+		if (prkey->params_len == sizeof(*keyinfo_gostparams))
+		{
+			keyinfo_gostparams = prkey->params;
+			sc_format_asn1_entry(asn1_gostr3410key_attr + 1,
+					&keyinfo_gostparams->gostr3410, NULL, 1);
+			sc_format_asn1_entry(asn1_gostr3410key_attr + 2,
+					&keyinfo_gostparams->gostr3411, NULL, 1);
+			sc_format_asn1_entry(asn1_gostr3410key_attr + 3,
+					&keyinfo_gostparams->gost28147, NULL, 1);
+		}
+		break;
 	default:
 		sc_error(ctx, "Invalid private key type: %X\n", obj->type);
 		SC_FUNC_RETURN(ctx, 0, SC_ERROR_INTERNAL);
@@ -404,6 +483,10 @@ sc_pkcs15_erase_prkey(struct sc_pkcs15_prkey *key)
 		free(key->u.dsa.g.data);
 		free(key->u.dsa.priv.data);
 		break;
+	case SC_ALGORITHM_GOSTR3410:
+		assert(key->u.gostr3410.d.data);
+		free(key->u.gostr3410.d.data);
+		break;
 	}
 	sc_mem_clear(key, sizeof(key));
 }
@@ -419,5 +502,7 @@ void sc_pkcs15_free_prkey_info(sc_pkcs15_prkey_info_t *key)
 {
 	if (key->subject)
 		free(key->subject);
+	if (key->params)
+		free(key->params);
 	free(key);
 }
