@@ -142,6 +142,16 @@ struct sc_pkcs15_prkey_dsa {
 	sc_pkcs15_bignum_t priv;
 };
 
+struct sc_pkcs15_pubkey_gostr3410 {
+	sc_pkcs15_bignum_t x;
+	sc_pkcs15_bignum_t y;
+};
+
+struct sc_pkcs15_prkey_gostr3410 {
+	/* private components */
+	sc_pkcs15_bignum_t d;
+};
+
 struct sc_pkcs15_pubkey {
 	int algorithm;
 
@@ -149,6 +159,7 @@ struct sc_pkcs15_pubkey {
 	union {
 		struct sc_pkcs15_pubkey_rsa rsa;
 		struct sc_pkcs15_pubkey_dsa dsa;
+		struct sc_pkcs15_pubkey_gostr3410 gostr3410;
 	} u;
 
 	/* DER encoded raw key */
@@ -161,6 +172,7 @@ struct sc_pkcs15_prkey {
 	union {
 		struct sc_pkcs15_prkey_rsa rsa;
 		struct sc_pkcs15_prkey_dsa dsa;
+		struct sc_pkcs15_prkey_gostr3410 gostr3410;
 	} u;
 };
 typedef struct sc_pkcs15_prkey sc_pkcs15_prkey_t;
@@ -241,6 +253,17 @@ typedef struct sc_pkcs15_data_info sc_pkcs15_data_info_t;
 #define SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE	0x08
 #define SC_PKCS15_PRKEY_ACCESS_LOCAL		0x10
 
+#define SC_PKCS15_PARAMSET_GOSTR3410_A          1
+#define SC_PKCS15_PARAMSET_GOSTR3410_B          2
+#define SC_PKCS15_PARAMSET_GOSTR3410_C          3
+
+#define SC_PKCS15_GOSTR3410_KEYSIZE             256
+
+struct sc_pkcs15_keyinfo_gostparams
+{
+	unsigned int gostr3410, gostr3411, gost28147;
+};
+
 struct sc_pkcs15_prkey_info {
 	struct sc_pkcs15_id id;	/* correlates to public certificate id */
 	unsigned int usage, access_flags;
@@ -248,6 +271,8 @@ struct sc_pkcs15_prkey_info {
 	size_t modulus_length;
 	u8     *subject;
 	size_t subject_len;
+	void   *params;
+	size_t params_len;
 
 	struct sc_path path;
 };
@@ -260,6 +285,8 @@ struct sc_pkcs15_pubkey_info {
 	size_t modulus_length;
 	u8     *subject;
 	size_t subject_len;
+	void   *params;
+	size_t params_len;
 
 	struct sc_path path;
 };
@@ -270,10 +297,12 @@ typedef struct sc_pkcs15_pubkey_info sc_pkcs15_pubkey_info_t;
 #define SC_PKCS15_TYPE_PRKEY			0x100
 #define SC_PKCS15_TYPE_PRKEY_RSA		0x101
 #define SC_PKCS15_TYPE_PRKEY_DSA		0x102
+#define SC_PKCS15_TYPE_PRKEY_GOSTR3410		0x103
 
 #define SC_PKCS15_TYPE_PUBKEY			0x200
 #define SC_PKCS15_TYPE_PUBKEY_RSA		0x201
 #define SC_PKCS15_TYPE_PUBKEY_DSA		0x202
+#define SC_PKCS15_TYPE_PUBKEY_GOSTR3410		0x203
 
 #define SC_PKCS15_TYPE_CERT			0x400
 #define SC_PKCS15_TYPE_CERT_X509		0x401
@@ -398,7 +427,9 @@ typedef struct sc_pkcs15_card {
 #define SC_PKCS15_CARD_FLAG_LOGIN_REQUIRED	0x02
 #define SC_PKCS15_CARD_FLAG_PRN_GENERATION	0x04
 #define SC_PKCS15_CARD_FLAG_EID_COMPLIANT	0x08
-#define SC_PKCS15_CARD_FLAG_SIGN_WITH_DECRYPT	0x10
+#define SC_PKCS15_CARD_FLAG_SIGN_WITH_DECRYPT	0x10000000
+#define SC_PKCS15_CARD_FLAG_EMULATED		0x20000000
+#define SC_PKCS15_CARD_FLAG_FIX_INTEGERS	0x40000000
 
 /* sc_pkcs15_bind:  Binds a card object to a PKCS #15 card object
  * and initializes a new PKCS #15 card object.  Will return
@@ -476,6 +507,10 @@ int sc_pkcs15_find_data_object_by_id(struct sc_pkcs15_card *p15card,
 int sc_pkcs15_find_data_object_by_app_oid(struct sc_pkcs15_card *p15card,
 					  const struct sc_object_id *app_oid,
 					  struct sc_pkcs15_object **out);
+int sc_pkcs15_find_data_object_by_name(struct sc_pkcs15_card *p15card,
+				const char *app_label,
+				const char *label,
+				struct sc_pkcs15_object **out);
 void sc_pkcs15_free_data_object(struct sc_pkcs15_data *data_object);
 
 int sc_pkcs15_read_certificate(struct sc_pkcs15_card *card,
@@ -662,6 +697,8 @@ typedef struct sc_pkcs15_search_key {
 
 	unsigned int		match_reference : 1;
 	int			reference;
+	const char *		app_label;
+	const char *		label;
 } sc_pkcs15_search_key_t;
 
 int sc_pkcs15_search_objects(sc_pkcs15_card_t *, sc_pkcs15_search_key_t *,
